/*
 *  cdc.c
 *
 *  Helper functions for the dynamic creation of Common Data Classes (CDCs)
 *
 *  Copyright 2014 Michael Zillgith
 *
 *  This file is part of libIEC61850.
 *
 *  libIEC61850 is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  libIEC61850 is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with libIEC61850.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  See COPYING file for the complete license text.
 */

#include "dynamic_model.h"
#include "cdc.h"

/************************************************
 * Constructed Attribute Classes
 ***********************************************/

DataAttribute*
CAC_AnalogueValue_create(char* name, ModelNode* parent, FunctionalConstraint fc, uint8_t triggerOptions,
        bool isIntegerNotFloat)
{
    DataAttribute* analogeValue = DataAttribute_create(name, parent, CONSTRUCTED, fc, triggerOptions, 0, 0);

    if (isIntegerNotFloat)
        DataAttribute_create("i", (ModelNode*) analogeValue, INT32, MX, triggerOptions, 0, 0);
    else
        DataAttribute_create("f", (ModelNode*) analogeValue, FLOAT32, MX, triggerOptions, 0, 0);


    return analogeValue;
}

/**
 * CDC_OPTION_AC_CLC_O
 */
DataAttribute*
CAC_Vector_create(char* name, ModelNode* parent, uint32_t options, FunctionalConstraint fc, uint8_t triggerOptions)
{
    DataAttribute* vector = DataAttribute_create(name, parent, CONSTRUCTED, fc, triggerOptions, 0, 0);

    CAC_AnalogueValue_create("mag", (ModelNode*) vector, fc, triggerOptions, false);

    if (options & CDC_OPTION_AC_CLC_O)
        CAC_AnalogueValue_create("ang", (ModelNode*) vector, fc, triggerOptions, false);

    return vector;
}

DataAttribute*
CAC_Point_create(char* name, ModelNode* parent, FunctionalConstraint fc, uint8_t triggerOptions, bool hasZVal)
{
    DataAttribute* point = DataAttribute_create(name, parent, CONSTRUCTED, fc, triggerOptions, 0, 0);

    DataAttribute_create("xVal", (ModelNode*) point, FLOAT32, fc, triggerOptions, 0, 0);
    DataAttribute_create("yVal", (ModelNode*) point, FLOAT32, fc, triggerOptions, 0, 0);

    if (hasZVal)
        DataAttribute_create("zVal", (ModelNode*) point, FLOAT32, fc, triggerOptions, 0, 0);

    return point;
}

DataAttribute*
CAC_ScaledValueConfig_create(char* name, ModelNode* parent)
{
    DataAttribute* scaling = DataAttribute_create(name, parent, CONSTRUCTED, CF, TRG_OPT_DATA_CHANGED, 0, 0);

    DataAttribute_create("scaleFactor", (ModelNode*) scaling, FLOAT32, CF, TRG_OPT_DATA_CHANGED, 0, 0);
    DataAttribute_create("offset", (ModelNode*) scaling, FLOAT32, CF, TRG_OPT_DATA_CHANGED, 0, 0);

    return scaling;
}

DataAttribute*
CAC_Unit_create(char* name, ModelNode* parent, bool hasMagnitude)
{
    DataAttribute* unit = DataAttribute_create(name, parent, CONSTRUCTED, CF, TRG_OPT_DATA_CHANGED, 0, 0);

    DataAttribute_create("SIUnit", (ModelNode*) unit, ENUMERATED, CF, TRG_OPT_DATA_CHANGED, 0, 0);

    if (hasMagnitude)
        DataAttribute_create("multiplier", (ModelNode*) unit, ENUMERATED, CF, TRG_OPT_DATA_CHANGED, 0, 0);

    return unit;
}

/************************************************
 * Common Data Classes - helper functions
 ***********************************************/

static void
CDC_addTimeQuality(DataObject* dataObject, FunctionalConstraint fc)
{
    DataAttribute_create("q", (ModelNode*) dataObject, QUALITY, fc, TRG_OPT_QUALITY_CHANGED, 0, 0);
    DataAttribute_create("t", (ModelNode*) dataObject, TIMESTAMP, fc, 0, 0, 0);
}

static void
CDC_addStatusToDataObject(DataObject* dataObject, DataAttributeType statusType)
{
    DataAttribute_create("stVal", (ModelNode*) dataObject, statusType, ST, TRG_OPT_DATA_CHANGED | TRG_OPT_DATA_UPDATE, 0, 0);
    CDC_addTimeQuality(dataObject, ST);
}

static void
CDC_addOptionPicsSubst(DataObject* dataObject, DataAttributeType type)
{
    DataAttribute_create("subEna", (ModelNode*) dataObject, BOOLEAN, SV, 0, 0, 0);
    DataAttribute_create("subVal", (ModelNode*) dataObject, type, SV, 0, 0, 0);
    DataAttribute_create("subQ", (ModelNode*) dataObject, QUALITY, SV, 0, 0, 0);
    DataAttribute_create("subID", (ModelNode*) dataObject, VISIBLE_STRING_64, SV, 0, 0, 0);
}

static void
CDC_addStandardOptions(DataObject* dataObject, uint32_t options)
{
    /* Standard options ? */
    if (options & CDC_OPTION_DESC)
        DataAttribute_create("d",(ModelNode*)  dataObject, VISIBLE_STRING_255, DC, 0, 0, 0);

    if (options & CDC_OPTION_DESC_UNICODE)
        DataAttribute_create("dU", (ModelNode*) dataObject, UNICODE_STRING_255, DC, 0, 0, 0);

    if (options & CDC_OPTION_AC_DLNDA) {
        DataAttribute_create("cdcNs", (ModelNode*) dataObject, VISIBLE_STRING_255, EX, 0, 0, 0);
        DataAttribute_create("cdcName", (ModelNode*) dataObject, VISIBLE_STRING_255, EX, 0, 0, 0);
    }

    if (options & CDC_OPTION_AC_DLN)
        DataAttribute_create("dataNs", (ModelNode*) dataObject, VISIBLE_STRING_255, EX, 0, 0, 0);
}

/************************************************
 * Common Data Classes - constructors
 ***********************************************/

DataObject*
CDC_SPS_create(char* dataObjectName, ModelNode* parent, uint32_t options)
{
    DataObject* newSPS = DataObject_create(dataObjectName, parent, 0);

    CDC_addStatusToDataObject(newSPS, BOOLEAN);

    if (options & CDC_OPTION_PICS_SUBST)
        CDC_addOptionPicsSubst(newSPS, BOOLEAN);

    if (options & CDC_OPTION_BLK_ENA)
        DataAttribute_create("blkEna", (ModelNode*) newSPS, BOOLEAN, BL, 0, 0, 0);

    CDC_addStandardOptions(newSPS, options);

    return newSPS;
}

DataObject*
CDC_DPS_create(char* dataObjectName, ModelNode* parent, uint32_t options)
{
    DataObject* newDPS = DataObject_create(dataObjectName, parent, 0);

    CDC_addStatusToDataObject(newDPS, CODEDENUM);

    if (options & CDC_OPTION_PICS_SUBST)
        CDC_addOptionPicsSubst(newDPS, CODEDENUM);

    if (options & CDC_OPTION_BLK_ENA)
        DataAttribute_create("blkEna", (ModelNode*) newDPS, BOOLEAN, BL, 0, 0, 0);

    CDC_addStandardOptions(newDPS, options);

    return newDPS;
}

DataObject*
CDC_INS_create(char* dataObjectName, ModelNode* parent, uint32_t options)
{
    DataObject* newINS = DataObject_create(dataObjectName, parent, 0);

    CDC_addStatusToDataObject(newINS, INT32);

    if (options & CDC_OPTION_PICS_SUBST)
        CDC_addOptionPicsSubst(newINS, INT32);

    if (options & CDC_OPTION_BLK_ENA)
        DataAttribute_create("blkEna", (ModelNode*) newINS, BOOLEAN, BL, 0, 0, 0);

    CDC_addStandardOptions(newINS, options);

    return newINS;
}


DataObject*
CDC_ENS_create(char* dataObjectName, ModelNode* parent, uint32_t options)
{
    DataObject* newENS = DataObject_create(dataObjectName, parent, 0);

    CDC_addStatusToDataObject(newENS, ENUMERATED);

    if (options & CDC_OPTION_PICS_SUBST)
        CDC_addOptionPicsSubst(newENS, ENUMERATED);

    if (options & CDC_OPTION_BLK_ENA)
        DataAttribute_create("blkEna", (ModelNode*) newENS, BOOLEAN, BL, 0, 0, 0);

    CDC_addStandardOptions(newENS, options);

    return newENS;
}

DataObject*
CDC_BCR_create(char* dataObjectName, ModelNode* parent, uint32_t options)
{
    DataObject* newBCR = DataObject_create(dataObjectName, parent, 0);

    DataAttribute_create("actVal", (ModelNode*) newBCR, INT64, ST, TRG_OPT_DATA_CHANGED, 0, 0);

    if (options & CDC_OPTION_FROZEN_VALUE) {
        DataAttribute_create("frVal", (ModelNode*) newBCR, INT64, ST, TRG_OPT_DATA_UPDATE, 0, 0);
        DataAttribute_create("frTm", (ModelNode*) newBCR, TIMESTAMP, ST, 0, 0, 0);
    }

    CDC_addTimeQuality(newBCR, ST);

    if (options & CDC_OPTION_UNIT)
        DataAttribute_create("units", (ModelNode*) newBCR, ENUMERATED, CF, TRG_OPT_DATA_CHANGED, 0, 0);

    DataAttribute_create("pulsQty", (ModelNode*) newBCR, FLOAT32, CF, TRG_OPT_DATA_CHANGED, 0, 0);

    if (options & CDC_OPTION_FROZEN_VALUE) {
        DataAttribute_create("frEna", (ModelNode*) newBCR, BOOLEAN, CF, TRG_OPT_DATA_CHANGED, 0, 0);
        DataAttribute_create("strTm", (ModelNode*) newBCR, TIMESTAMP, CF, TRG_OPT_DATA_CHANGED, 0, 0);
        DataAttribute_create("frPd", (ModelNode*) newBCR, INT32, CF, TRG_OPT_DATA_CHANGED, 0, 0);
        DataAttribute_create("frRs", (ModelNode*) newBCR, BOOLEAN, CF, TRG_OPT_DATA_CHANGED, 0, 0);
    }

    CDC_addStandardOptions(newBCR, options);

    return newBCR;
}

DataObject*
CDC_SEC_create(char* dataObjectName, ModelNode* parent, uint32_t options)
{
    DataObject* newSEC = DataObject_create(dataObjectName, parent, 0);

    DataAttribute_create("cnt", (ModelNode*) newSEC, INT32U, ST, TRG_OPT_DATA_CHANGED, 0, 0);
    DataAttribute_create("sev", (ModelNode*) newSEC, ENUMERATED, ST, 0, 0, 0);
    DataAttribute_create("t", (ModelNode*) newSEC, TIMESTAMP, ST, 0, 0, 0);

    if (options & CDC_OPTION_ADDR)
        DataAttribute_create("addr", (ModelNode*) newSEC, OCTET_STRING_64, ST, 0, 0, 0);

    if (options & CDC_OPTION_ADDINFO)
        DataAttribute_create("addInfo", (ModelNode*) newSEC, VISIBLE_STRING_64, ST, 0, 0, 0);

    CDC_addStandardOptions(newSEC, options);

    return newSEC;
}




/**
 * CDC_OPTION_INST_MAG
 * CDC_OPTION_RANGE
 * CDC_OPTION_PICS_SUBST
 */
DataObject*
CDC_MV_create(char* dataObjectName, ModelNode* parent, uint32_t options, bool isIntegerNotFloat)
{
    DataObject* newMV = DataObject_create(dataObjectName, parent, 0);

    if (options & CDC_OPTION_INST_MAG)
        CAC_AnalogueValue_create("instMag", (ModelNode*) newMV, MX, 0, isIntegerNotFloat);

    CAC_AnalogueValue_create("mag", (ModelNode*) newMV, MX, TRG_OPT_DATA_CHANGED | TRG_OPT_DATA_UPDATE, isIntegerNotFloat);

    if (options & CDC_OPTION_RANGE)
        DataAttribute_create("range", (ModelNode*) newMV, ENUMERATED, MX, TRG_OPT_DATA_CHANGED, 0, 0);

    CDC_addTimeQuality(newMV, MX);

//    if (options & CDC_OPTION_PICS_SUBST)
//        CDC_addOptionPicsSubst(newMV, )

    CDC_addStandardOptions(newMV, options);

    return newMV;
}

/**
 * CDC_OPTION_INST_MAG
 * CDC_OPTION_RANGE
 */
DataObject*
CDC_CMV_create(char* dataObjectName, ModelNode* parent, uint32_t options)
{
    DataObject* newMV = DataObject_create(dataObjectName, parent, 0);

    if (options & CDC_OPTION_INST_MAG)
        CAC_Vector_create("instCVal", (ModelNode*) newMV, options, MX, 0);

    CAC_Vector_create("cVal", (ModelNode*) newMV, options, MX, TRG_OPT_DATA_CHANGED | TRG_OPT_DATA_UPDATE);

    if (options & CDC_OPTION_RANGE)
        DataAttribute_create("range", (ModelNode*) newMV, ENUMERATED, MX, TRG_OPT_DATA_CHANGED, 0, 0);

    if (options & CDC_OPTION_RANGE_ANG)
           DataAttribute_create("rangeAng", (ModelNode*) newMV, ENUMERATED, MX, TRG_OPT_DATA_CHANGED, 0, 0);

    CDC_addTimeQuality(newMV, MX);

//    if (options & CDC_OPTION_PICS_SUBST)
//        CDC_addOptionPicsSubst(newMV, )

    CDC_addStandardOptions(newMV, options);

    return newMV;
}


/**
 * CDC_OPTION_UNIT
 * CDC_OPTION_AC_SCAV
 * CDC_OPTION_MIN
 * CDC_OPTION_MAX
 */
DataObject*
CDC_SAV_create(char* dataObjectName, ModelNode* parent, uint32_t options, bool isIntegerNotFloat)
{
    DataObject* newSAV = DataObject_create(dataObjectName, parent, 0);

    CAC_AnalogueValue_create("instMag", (ModelNode*) newSAV, MX, 0, isIntegerNotFloat);

    CDC_addTimeQuality(newSAV, MX);

    if (options & CDC_OPTION_UNIT)
        CAC_Unit_create("units", (ModelNode*) newSAV, options & CDC_OPTION_UNIT_MULTIPLIER);

    if (options & CDC_OPTION_AC_SCAV)
        CAC_ScaledValueConfig_create("sVC", (ModelNode*) newSAV);

    if (options & CDC_OPTION_MIN)
        CAC_AnalogueValue_create("min", (ModelNode*) newSAV, CF, TRG_OPT_DATA_CHANGED, isIntegerNotFloat);

    if (options & CDC_OPTION_MAX)
        CAC_AnalogueValue_create("max", (ModelNode*) newSAV, CF, TRG_OPT_DATA_CHANGED, isIntegerNotFloat);

    CDC_addStandardOptions(newSAV, options);

    return newSAV;
}
